/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2007-2011 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdlib.h>
#include <GL/gl.h>
#include <errno.h>

#include "gl/u2d_DisplaySet.h"
#include "gl/u2d_DisplayContainer.h"
#include "gl/u2d_DisplayText.h"
#include "common/exceptions.h"
#include "common/u2d_Mesh.h"
#include "solver/u2d_Problem.h"
#include "u2d_Math.h"

#ifndef glBindBuffer
void glBindBuffer(GLenum /*target*/, GLuint /*buffer*/) { };
void glGenBuffers(GLsizei /*n*/, GLuint * /*buffers*/) {};
void glDeleteBuffers(GLsizei /*n*/, const GLuint * /*buffers*/) {};
void glBufferData(GLenum /*target*/, GLsizeiptr /*size*/, const GLvoid * /*data*/, GLenum /*usage*/) {};
#endif

u2d_DisplaySet::u2d_DisplaySet(string _name /* = "" */)
        throw (u2d_OpenGL_Exception &) :
vertexType(0),
name(_name),
compiledDL(false),
compiledVBO(false),
drawing(false),
enabled(true),
fillMode(GL_LINE),
lineWidth(1),
lineColor({1.0, 1.0, 1.0}), // C++0x
coordPerVertex(0),
vertexPerElement(0),
colorPerVertex(3),
displayList(0),
VBOVertexBuffer(0),
VBOTopologyBuffer(0) {

}

u2d_DisplaySet::~u2d_DisplaySet() {
    this->destroyDL();
    this->destroyVBO();
}

VERTEX_PRECISION *u2d_DisplaySet::getVertexArray() {
    return &vertexArray.front();
}

void u2d_DisplaySet::pushVertexCoord(VERTEX_PRECISION a) {
    vertexArray.push_back(a);
}

void u2d_DisplaySet::pushVertexCoord(VERTEX_PRECISION a, VERTEX_PRECISION b) {
    vertexArray.push_back(a);
    vertexArray.push_back(b);
}

void u2d_DisplaySet::pushVertexCoord(VERTEX_PRECISION a, VERTEX_PRECISION b, VERTEX_PRECISION c) {
    vertexArray.push_back(a);
    vertexArray.push_back(b);
    vertexArray.push_back(c);
}

/**
 * Sets the number of coordinates for each vertex. I.e. 2 for 2D plots, 3 for
 * 3D plots. 
 * @param m_coord_per_vertex The number of coordinates per vertex, 2 or 3.
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setCoordPerVertex(GLint m_coord_per_vertex) {
    coordPerVertex = m_coord_per_vertex;
    return *this;
}

GLint u2d_DisplaySet::getCoordPerVertex() const {
    return coordPerVertex;
}

GLsizei u2d_DisplaySet::getVertexCoordCount() const {
    return vertexArray.size();
}

GLsizei u2d_DisplaySet::getVertexCount() const {
    if(coordPerVertex == 0) return 0;
    return vertexArray.size()/coordPerVertex;
}

u2d_DisplaySet& u2d_DisplaySet::setVertexType(GLenum _vertexType) throw (u2d_OpenGL_Exception &) {
    if ((_vertexType == GL_LINES) ||
            (_vertexType == GL_TRIANGLES) ||
            (_vertexType == GL_QUADS)) {
        this->vertexType = _vertexType;
    } else {
        throw u2d_OpenGL_Exception("Unsupported vertex type");
    }
    return *this;
}

GLenum u2d_DisplaySet::getVertexType() const {
    return this->vertexType;
}
    
/**
 * @remark Newer C++ standards guarantee that the vector from which we obtain
 * the pointer we return use contiguous memory, so the array returned should
 * be consistent UNTIL you modify the set.
 * @return A pointer to an array with topology indices.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
GLuint *u2d_DisplaySet::getTopologyArray() {
    return &topologyArray.front();
}
/**
 * The number of vertexes that
 * @param m_coord_per_topology
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setVertexPerElement(GLint m_coord_per_topology) {
    vertexPerElement = m_coord_per_topology;
    return *this;
}

GLint u2d_DisplaySet::getVertexPerElement() const {
    return vertexPerElement;
}

GLsizei u2d_DisplaySet::getElementCount() const {
    if(vertexPerElement == 0) return 0;
    return topologyArray.size()/vertexPerElement;
}

GLsizei u2d_DisplaySet::getTopologyElementCount() const {
    return topologyArray.size();
}

void u2d_DisplaySet::pushTopologyElement(GLuint a) {
    topologyArray.push_back(a);
}

void u2d_DisplaySet::pushTopologyElement(GLuint a, GLuint b) {
    topologyArray.push_back(a);
    topologyArray.push_back(b);
}

void u2d_DisplaySet::pushTopologyElement(GLuint a, GLuint b, GLuint c) {
    topologyArray.push_back(a);
    topologyArray.push_back(b);
    topologyArray.push_back(c);
}

/**
 * Sets the number of colors per vertex. Allowed values are 3 or 4 and are
 * imposed by the underlying OpenGL implementations. Values are intended as
 * red green blue plus alpha channel if 4 is specified.
 * @param _colorsPerVertex The number of colors per vertex, either 3 or 4
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setColorsPerVertex(GLint _colorsPerVertex) throw (u2d_OpenGL_Exception &) {
    if ((_colorsPerVertex == 3) ||
            (_colorsPerVertex == 4)) {
        this->colorPerVertex = _colorsPerVertex;
    } else {
        throw u2d_OpenGL_Exception("Unsupported color number");
    }
    return *this;
}

/**
 * Gets the set number of colors per vertex. Possible return values are 3 or 4.
 * @return The number of colors per vertex, 3 or 4.
 */
GLint u2d_DisplaySet::getColorsPerVertex() const {
    return this->colorPerVertex;
}

/**
 * Adds a color to the color vector. This color is intended specified for the
 * vertex in the corresponding position of the vertex vector.
 * @param a red value
 * @param b green value
 * @param c blue value
 */
void u2d_DisplaySet::pushVertexColor(GLfloat a, GLfloat b, GLfloat c) {
    colorArray.push_back(a);
    colorArray.push_back(b);
    colorArray.push_back(c);
}

/**
 * Gets the number of colors in the color vector. A color in intended as a
 * triplet of rgb values, plus eventually the alpha value.
 * @return The number of colors in the array. That is the number of elements
 * in the array divided by the number of values to make up a color (may be 3
 * or 4 depending on what has been set with setColorsPerVertex() ).
 */
GLsizei u2d_DisplaySet::getColorCount() const {
    if(colorPerVertex == 0) return 0;
    return colorArray.size()/colorPerVertex;    
}

/**
 * The size of the color array.
 * @return The size of the color array
 */
GLsizei u2d_DisplaySet::getColorValuesCount() const {
    return this->colorArray.size();
}

/**
 * Returns the color array. The pointer is valid till this class is valid as it
 * points to class internal storage. It may become invalid if any push
 * operation is performed on the color vector.
 * @return A pointer to the color array.
 */
GLfloat *u2d_DisplaySet::getColorArray() {
    return &colorArray.front();
}

/**
 * Change the fill mode of drawn polygons. The change is performed for both
 * front and back face.
 * Since fillMode is a status for OpenGL that get reseted after each frame,
 * there is no need to recompile DL and VBO since it's not saved in the DL or
 * somewhere else.
 * Default value is GL_LINE.
 * @param fillMode Valid values are GL_LINE, GL_FILL and GL_POINT.
 * From OpenGL documentation:
 * GL_POINT Polygon vertices that are marked as the start of a boundary edge
 * are drawn as points. Point attributes such as GL_POINT_SIZE and
 * GL_POINT_SMOOTH control the rasterization of the points.
 * Polygon rasterization attributes other than GL_POLYGON_MODE have no effect.
 * GL_LINE Boundary edges of the polygon are drawn as line segments.
 * They are treated as connected line segments for line stippling;
 * the line stipple counter and pattern are not reset between segments
 * (see glLineStipple). Line attributes such as GL_LINE_WIDTH and
 * GL_LINE_SMOOTH control the rasterization of the lines.
 * Polygon rasterization attributes other than GL_POLYGON_MODE have no effect.
 * GL_FILL The interior of the polygon is filled. Polygon attributes such as
 * GL_POLYGON_STIPPLE and GL_POLYGON_SMOOTH control the rasterization
 *  of the polygon.
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setFillMode(GLenum _fillMode) {
    if ((_fillMode != GL_LINE) &&
            (_fillMode != GL_FILL) &&
            (_fillMode != GL_POINT)) {
        throw u2d_OpenGL_Exception("Unsupported fill mode");
    }
    /* Change the mode */
    this->fillMode = _fillMode;
    return *this;
}

/**
 * Sets the line width for drawing this object contents.
 * Default value is 1.
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setLineWidth(GLfloat _lineWidth) {
    this->lineWidth = _lineWidth;
    return *this;
}

/**
 * Sets the line color for drawing this object contents. This color is used
 * in case no color is specified via the color array.
 * Allowed ranged for each value are between 0.0 and 1.0 and correspond to
 * RGB values.
 * Default value is (1, 1, 1) that is white.
 * @param red
 * @param green
 * @param blue
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::setLineColor(GLfloat _red, GLfloat _green, GLfloat _blue) {
    this->lineColor[0] = _red;
    this->lineColor[1] = _green;
    this->lineColor[2] = _blue;
    return *this;
}

/**
 * Enable object for drawing.
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::enable() {
    this->enabled = true;
    return *this;
}

/**
 * Disable object for drawing.
 * @return A reference to this object useful for chaining other methods.
 */
u2d_DisplaySet& u2d_DisplaySet::disable() {
    this->enabled = false;
    return *this;
}

/**
 * Checks if the object is enabled for drawing or not.
 * @return True if object is enabled for drawing, false otherwise.
 */
bool u2d_DisplaySet::isEnabled() const {
    return this->enabled;
}

/**
 * 
 * drawing is set for the duration of the drawing to be
 * sure the vector does not reallocate the array and invalidates the pointer.
 */
void u2d_DisplaySet::draw() {
    if (isEnabled()) {
        /* Flag the list as drawing so to stop vertex addition and avoid invalidation
         * of vertex array and topology arrays pointers.
         * This does not prevent another thread from calling draw(),
         * but this should simply have no effect cause the display list is still
         * empty or contains the old list. */
        this->drawing = true;
        drawDL();
        //std::cout << "Draw: " << this->getName() << std::endl;
        this->drawing = false;
    }
}

/**
 * Draws the data using Vertex Array.
 * This does not need to set final to the class definitely, the vertex and
 * topology data may change since they are transmitted each frame to the OpenGL
 * implementation. 
 * Check to use this in the future: EXT_compiled_vertex_array.
 */        
void u2d_DisplaySet::drawVA() {

    /* Do not load and lose time on empty lists */
    if(this->getVertexCount() == 0) {
        return;
    }
    
    /* Set the fill mode*/
    glPolygonMode(GL_FRONT_AND_BACK, fillMode);
    
    /* Set the line width */
    glLineWidth(this->lineWidth);
    
    /* Set line basic color */
    glColor3fv(this->lineColor);
    
    /* Enable vertex arrays */
    glEnableClientState(GL_VERTEX_ARRAY);

    /* Add the vertexes */
    glVertexPointer(this->getCoordPerVertex(),
            VERTEX_PRECISION_TYPE, 0, (GLvoid *) this->getVertexArray());

    /* Add the colors */
    if (this->getColorCount() != 0) {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(this->getColorsPerVertex(),
                GL_FLOAT, 0, (GLvoid *) this->getColorArray());
    }
    
    /* Switch based on vertex type:
     * - If we have GL_LINES and vertexPerElement == 2 and topology is not
     *   zero length we draw segments between specified topology vertexes
     *   as defined by the couples of vertex indexes given by the topology.
     * - If we have GL_LINES and topology is zero length we draw lines
     *   between consecutive couples of segments and  _ignore_ the topology.
     * - If we have GL_TRIANGLES and vertexPerElement == 3 and topology is not
     *   zero length we draw triangles as defined by the triplets of vertex
     *   indexes given by the topology 
     * - If we have GL_TRIANGLES and topology is zero length we draw triangles
     *   with consecutive triplets of vertexes and _ignore_ the topology */
    if((this->vertexType == GL_LINES) &&
            (vertexPerElement == 2) && 
            (this->getTopologyElementCount() != 0)) {
        
        glDrawElements(GL_LINES, this->getTopologyElementCount(),\
                       GL_UNSIGNED_INT, this->getTopologyArray());
        
    } else if((this->vertexType == GL_LINES)) {
        
        glDrawArrays(GL_LINES, 0, this->getVertexCount());

    } else if((this->vertexType == GL_TRIANGLES) &&
            (vertexPerElement == 3) &&
            (this->getTopologyElementCount() != 0)) {
        
        glDrawElements(GL_TRIANGLES, this->getTopologyElementCount(),\
                       GL_UNSIGNED_INT, this->getTopologyArray());
        
    } else if(this->vertexType == GL_TRIANGLES) {
        
        glDrawArrays(GL_TRIANGLES, 0, this->getVertexCount());
        
    } else if(this->vertexType == GL_QUADS) {
        
         glDrawArrays(GL_QUADS, 0, this->getVertexCount());
         
    } else {
        
        throw u2d_OpenGL_Exception("Inconsistency between vertex type and topology specification.");
        
    }
    
    /* Disable the client state*/
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    
}

/**
 *  Draws the data using Display List.
 * 
 */        
void u2d_DisplaySet::drawDL() {
    if(!this->compiledDL) {
        this->compileDL();
    }
    
    /* Set the fill mode*/
    glPolygonMode(GL_FRONT_AND_BACK, fillMode);
    
    /* Set the line width */
    glLineWidth(this->lineWidth);
    
    /* Set line basic color */
    glColor3fv(this->lineColor);
    
    /* Nothing to do here, we just need to call and we are done.
     * this is because in the case of display lists all the structure
     * has been already assembled and transferred with compileDL().*/
    glCallList(this->displayList);
}

/**
 *  Draws the data using Video Buffer Object.
 * 
 */        
void u2d_DisplaySet::drawVBO() {
    if(!this->compiledVBO) {
        this->compileVBO();
    }
    
    /* Set the fill mode*/
    glPolygonMode(GL_FRONT_AND_BACK, fillMode);
    
    /* Set the line width */
    glLineWidth(this->lineWidth);
    
    /* Set line basic color */
    glColor3fv(this->lineColor);
    
    /* Bind VBOs for vertex array and index array */
    glBindBuffer(GL_ARRAY_BUFFER, VBOVertexBuffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, VBOTopologyBuffer);

    /* Enable vertex array mode */
    glEnableClientState(GL_VERTEX_ARRAY);
   
    /* Add the vertexes.
     * Notice that in this case the last parameter is the offset in the
     * VBO buffer and not a pointer */
    glVertexPointer(this->getCoordPerVertex(), \
                    VERTEX_PRECISION_TYPE, 0, 0);
    
    /* Switch based on vertex type:
     * - If we have GL_LINES and vertexPerElement == 2 and topology is not
     *   zero length we draw segments between specified topology vertexes
     *   as defined by the couples of vertex indexes given by the topology.
     * - If we have GL_LINES and topology is zero length we draw lines
     *   between consecutive couples of segments and  _ignore_ the topology.
     * - If we have GL_TRIANGLES and vertexPerElement == 3 and topology is not
     *   zero length we draw triangles as defined by the triplets of vertex
     *   indexes given by the topology 
     * - If we have GL_TRIANGLES and topology is zero length we draw triangles
     *   with consecutive triplets of vertexes and _ignore_ the topology */
    if((this->vertexType == GL_LINES) &&
            (vertexPerElement == 2) && 
            (this->getTopologyElementCount() != 0)) {
        
        glDrawElements(GL_LINES, this->getTopologyElementCount(),\
                       GL_UNSIGNED_INT, 0);
        
    } else if((this->vertexType == GL_LINES)) {
        
        glDrawArrays(GL_LINES, 0, this->getVertexCount());
        
    } else if((this->vertexType == GL_TRIANGLES) &&
            (vertexPerElement == 3) &&
            (this->getTopologyElementCount() != 0)) {
        
        glDrawElements(GL_TRIANGLES, this->getTopologyElementCount(),\
                       GL_UNSIGNED_INT, 0);
        
    } else if((this->vertexType == GL_TRIANGLES)
            ) {
        
        glDrawArrays(GL_TRIANGLES, 0, this->getVertexCount());
        
    } else if(this->vertexType == GL_QUADS) {
        
         glDrawArrays(GL_QUADS, 0, this->getVertexCount());
         
    } else {
        
        throw u2d_OpenGL_Exception("Inconsistency between vertex type and topology specification.");
        
    }
    
    glDisableClientState(GL_VERTEX_ARRAY); // deactivate vertex array
    
    /* Bind back with 0, so, switch back to normal pointer operation */
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    
}
/**
 * Get the list ready for calling. After the list has been readied you cannot
 * anymore add more vertices to it. On the contrary to call() a list it must
 * be ready else the call() method will have no effect.
 * This function actually compiles the list and moves it to the card memory.
 * To modify the list you need to unReady() it, modify and then getReady() it
 * again.
 * This are the rules used to decide what to draw with the given data:
 * - If we have GL_LINES and coordPerElement less or equal 1 we draw lines and
 *   _ignore_ the topology.
 * - If we have GL_Lines and coordPerElement == 2 we draw segments
 *   as defined by the couples of vertex indexes given by the topology.
 * - If we have GL_TRIANGLES and coordPerElement == 3 we draw triangles
 *   as defined by the triplets of vertex indexes given by the topology
 * @throw u2d_OpenGL_Exception if something wrong happens. E.g. The data type
 * is inconsistent.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_DisplaySet::compileDL()
throw (u2d_OpenGL_Exception &) {
    /* Create a DL for compiling data. */
    this->createDL();

    /* Flag the list as ready so to stop vertex addition and avoid invalidation
     * of vertex array and topology arrays pointers.
     * This does not prevent another thread from calling draw(),
     * but this should simply have no effect cause the display list is still
     * empty or contains the old list. */
    this->compiledDL = true;
    
    /* Do not load and lose time on empty lists */
    if(this->getVertexCount() == 0) {
        return;
    }
 
    /* Select the list we want to use.
     * We just want to compile it and execute it later */
    glNewList(this->displayList, GL_COMPILE);
    
    /* Switch based on vertex type:
     * - If we have GL_LINES and vertexPerElement == 2 and topology is not
     *   zero length we draw segments between specified topology vertexes
     *   as defined by the couples of vertex indexes given by the topology.
     * - If we have GL_LINES and topology is zero length we draw lines
     *   between consecutive couples of segments and  _ignore_ the topology.
     * - If we have GL_TRIANGLES and vertexPerElement == 3 and topology is not
     *   zero length we draw triangles as defined by the triplets of vertex
     *   indexes given by the topology 
     * - If we have GL_TRIANGLES and topology is zero length we draw triangles
     *   with consecutive triplets of vertexes and _ignore_ the topology */
    if((this->vertexType == GL_LINES) &&
            (vertexPerElement == 2) && 
            (this->getTopologyElementCount() != 0)) {
        
        GLuint idxCoordX = 0;
        GLuint idxCoordY = 0;
        size_t colorIndex = 0;

        glBegin(GL_LINES);
        std::vector<GLuint>::iterator iter = topologyArray.begin();
        while(iter != this->topologyArray.end()) {
            /* Add the color */
            if (this->getColorCount() != 0) {
                colorIndex = colorPerVertex * *iter; 
                glColor3f(this->colorArray[colorIndex], this->colorArray[colorIndex + 1], this->colorArray[colorIndex + 2]);
            }
            
            /* Add the vertex */
            idxCoordX = coordPerVertex * *iter;
            idxCoordY = coordPerVertex * *iter + 1;
            if(VERTEX_PRECISION_TYPE == GL_FLOAT) {
                glVertex2f(vertexArray[idxCoordX], vertexArray[idxCoordY]);
            } else if(VERTEX_PRECISION_TYPE == GL_DOUBLE) {
                glVertex2d(vertexArray[idxCoordX], vertexArray[idxCoordY]);
            }
            ++iter;
        }
        glEnd();

    } else if((this->vertexType == GL_LINES)) {

        size_t colorIndex = 0;
        
        glBegin(GL_LINES);
        std::vector<VERTEX_PRECISION>::iterator iter = vertexArray.begin();
        while(iter != this->vertexArray.end()) {

            /* Add the color */
            if (this->getColorCount() != 0) {
                glColor3f(this->colorArray[colorIndex], this->colorArray[colorIndex + 1], this->colorArray[colorIndex + 2]);
            }
            colorIndex += 3;
            
            /* Add the vertex */
            if (VERTEX_PRECISION_TYPE == GL_FLOAT) {
                glVertex2f(*iter, *(iter + 1));
            } else if(VERTEX_PRECISION_TYPE == GL_DOUBLE) {
                glVertex2d(*iter, *(iter + 1));
            }
            ++iter;
            ++iter;
        }
        glEnd();
        

    } else if((this->vertexType == GL_TRIANGLES) &&
            (vertexPerElement == 3) &&
            (this->getTopologyElementCount() != 0)) {
        
        GLuint idxCoordX = 0;
        GLuint idxCoordY = 0;
        size_t colorIndex = 0;

        glBegin(GL_TRIANGLES);
        std::vector<GLuint>::iterator iter = topologyArray.begin();
        while(iter != this->topologyArray.end()) {
            /* Add the color */
            if (this->getColorCount() != 0) {
                colorIndex = colorPerVertex * *iter; 
                glColor3f(this->colorArray[colorIndex], this->colorArray[colorIndex + 1], this->colorArray[colorIndex + 2]);
            }
            
            /* Add the vertex */
            idxCoordX = coordPerVertex * *iter;
            idxCoordY = coordPerVertex * *iter + 1;
            if(VERTEX_PRECISION_TYPE == GL_FLOAT) {
                glVertex2f(vertexArray[idxCoordX], vertexArray[idxCoordY]);
            } else if(VERTEX_PRECISION_TYPE == GL_DOUBLE) {
                glVertex2d(vertexArray[idxCoordX], vertexArray[idxCoordY]);
            }
            
            ++iter;
        }
        glEnd();
        
    
    } else if(this->vertexType == GL_TRIANGLES) {
        
        size_t colorIndex = 0;
        
        glBegin(GL_TRIANGLES);
        std::vector<VERTEX_PRECISION>::iterator iter = vertexArray.begin();
        while(iter != this->vertexArray.end()) {
            /* Add the color */
            if (this->getColorCount() != 0) {
                glColor3f(this->colorArray[colorIndex], this->colorArray[colorIndex + 1], this->colorArray[colorIndex + 2]);
            }
            colorIndex += 3;
            
            /* Add the vertex */
            if(VERTEX_PRECISION_TYPE == GL_FLOAT) {
                glVertex2f(*iter, *(iter + 1));
            } else if (VERTEX_PRECISION_TYPE == GL_DOUBLE) {
                glVertex2d(*iter, *(iter + 1));
            }
            ++iter;
            ++iter;
        }
        glEnd();

    
    } else if(this->vertexType == GL_QUADS) {
        
        size_t colorIndex = 0;
        
        glBegin(GL_QUADS);
        std::vector<VERTEX_PRECISION>::iterator iter = vertexArray.begin();
        while(iter != this->vertexArray.end()) {
            /* Add the color */
            if (this->getColorCount() != 0) {
                glColor3f(this->colorArray[colorIndex], this->colorArray[colorIndex + 1], this->colorArray[colorIndex + 2]);
            }
            colorIndex += 3;
            
            /* Add the vertex */
            if(VERTEX_PRECISION_TYPE == GL_FLOAT) {
                glVertex2f(*iter, *(iter + 1));
            } else if (VERTEX_PRECISION_TYPE == GL_DOUBLE) {
                glVertex2d(*iter, *(iter + 1));
            }
            ++iter;
            ++iter;
        }
        glEnd();

    
    } else {       
    
        throw u2d_OpenGL_Exception("Inconsistency between vertex type and topology specification.");
        
    }
    
    /* Close the list */
    glEndList();
}

void u2d_DisplaySet::compileVBO() throw (u2d_OpenGL_Exception &) {
    /* Create a VBO for compiling data. */
    this->createVBO();

    /* Usage flags:
     * "Static" means the data in VBO will not be changed (specified once and
     * used many times), "dynamic" means the data will be changed frequently
     * (specified and used repeatedly), and "stream" means the data will be
     * changed every frame (specified once and used once). "Draw" means the
     * data will be sent to GPU in order to draw (application to GL), "read"
     * means the data will be read by the client's application (GL to
     * application), and "copy" means the data will be used both drawing and
     * reading (GL to GL).
     * GL_STATIC_DRAW
     * GL_STATIC_READ
     * GL_STATIC_COPY
     * GL_DYNAMIC_DRAW
     * GL_DYNAMIC_READ
     * GL_DYNAMIC_COPY
     * GL_STREAM_DRAW
     * GL_STREAM_READ
     * GL_STREAM_COPY
     * Note that only draw token is useful for VBO, and copy and read token
     * will be become meaningful only for pixel/frame buffer object (PBO or FBO). 
     */

    glBufferData(GL_ARRAY_BUFFER,
            this->getVertexCount() * sizeof (VERTEX_PRECISION),
            (GLvoid *) this->getVertexArray(),
            GL_STATIC_DRAW);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,
            this->getTopologyElementCount() * sizeof (GLuint),
            (GLvoid *) this->getTopologyArray(),
            GL_STATIC_DRAW);

    this->compiledVBO = true;
}
 
u2d_DisplaySet& u2d_DisplaySet::setName(std::string _name) {
     this->name = _name;
     return *this;
}

string u2d_DisplaySet::getName() const {
    return this->name;
}

/**
 * Requests to the underlying OpenGL implementation a display list to use
 * for this display set.
 * @throw u2d_OpenGL_Exception in case of failure.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_DisplaySet::createDL()
throw (u2d_OpenGL_Exception &) {
    /* If we already have a list, delete it and create a new one*/
    this->destroyDL();

    /* Generate a list for this display set.*/
    this->displayList = glGenLists(1);
    /* If the list is still 0 or it's invalid there was an error */
    if(this->displayList == 0 || (glIsList(this->displayList) != GL_TRUE)) {
        throw u2d_OpenGL_Exception("Cannot get a display list for OpenGL drawing");
    }
}

void u2d_DisplaySet::destroyDL() {
    if (this->displayList != 0) {
        glDeleteLists(this->displayList, 1);
    }
}

void u2d_DisplaySet::createVBO()
throw (u2d_OpenGL_Exception &) {
    /* If we already have a list, delete it and create a new one*/
    this->destroyVBO();
    
    /* Create the VBO vertex buffer*/
    glGenBuffers(1, &VBOVertexBuffer);
    /* Create the VBO topology buffer*/
    glGenBuffers(1, &VBOTopologyBuffer);

}

void u2d_DisplaySet::destroyVBO() {
    if (this->VBOVertexBuffer != 0) {
        glDeleteBuffers(1, &VBOVertexBuffer);
    }
    if (this->VBOTopologyBuffer != 0) {
        glDeleteBuffers(1, &VBOTopologyBuffer);
    }
}

/**
 * Returns a u2d_DisplaySet useful for drawing to OpenGL the given mesh.
 * This function will prepare an object to show the domain part of the
 * given mesh.
 * @param mesh the mesh object you want to draw.
 * @return A display set for drawing the given mesh.
 */
u2d_DisplaySet *u2d_DisplaySet::FromMeshDomain(u2d_Mesh* mesh) {
     u2d_DisplaySet *displaySet = new u2d_DisplaySet();
    
    /* Vertexes */
    displaySet->setCoordPerVertex(mesh->getSpatialDimension());
    
    for(size_t i = 0; i < mesh->getDomainNodeNumber(); i++) {
        displaySet->pushVertexCoord(mesh->rr(0, i), mesh->rr(1, i));
    }
    
    /* Connectivity */
    displaySet->setVertexPerElement(3)
            .setVertexType(GL_TRIANGLES);
    
    for(size_t i = 0; i < mesh->getDomainElementNumber(); i++) {
        displaySet->pushTopologyElement(mesh->jjd(0, i) - 1,
                mesh->jjd(1, i) - 1, 
                mesh->jjd(2, i) - 1);
    }
        
    return displaySet;
}

/**
 * Returns a u2d_DisplaySet useful for drawing to OpenGL the given mesh.
 * This function will prepare an object to show the domain part of the
 * given mesh.
 * @param mesh the mesh object you want to draw.
 * @return A display set for drawing the given mesh.
 */
u2d_DisplayContainer *u2d_DisplaySet::FromMeshDomainWithSolution(u2d_Mesh *mesh, u2d_Problem *problem) {
     u2d_DisplaySet *displaySet = new u2d_DisplaySet();
    
    /* Vertexes */
    displaySet->setCoordPerVertex(mesh->getSpatialDimension());
    
    for(size_t i = 0; i < mesh->getDomainNodeNumber(); i++) {
        displaySet->pushVertexCoord(mesh->rr(0, i), mesh->rr(1, i));
    }
    
    /* Connectivity */
    displaySet->setVertexPerElement(3)
            .setVertexType(GL_TRIANGLES)
            .setFillMode(GL_FILL);
    
    for(size_t i = 0; i < mesh->getDomainElementNumber(); i++) {
        displaySet->pushTopologyElement(mesh->jjd(0, i) - 1,
                mesh->jjd(1, i) - 1, 
                mesh->jjd(2, i) - 1);
    }
    
    /* Colors */
    displaySet->setColorsPerVertex(3);
    GLfloat color[3];
    dense_vector<double> solution = problem->getSolution();
    float solution_min = mtl::min(solution);
    float solution_max = mtl::max(solution);
    //std::cout << solution_min << " <-min max->" << solution_max << std::endl;
    for(size_t i = 0; i < problem->getSize(); i++) {
        u2d_Math::LinearRGB(color, solution(i), solution_min, solution_max);
        displaySet->pushVertexColor(color[0], color[1], color[2]);
        
        //std::cout << solution(i) << " <-sol color->" << color[0] << " " << color[1] << " " << color[2] << std::endl;
        
    }
        
    /* Create the container and add the children */
    u2d_DisplayContainer *displayContainer = new u2d_DisplayContainer();
    displayContainer->addChild(displaySet);
    /* Generate the mesh to overlap to the solution */
    displayContainer->addChild(u2d_DisplaySet::FromMeshDomain(mesh));
            
    return displayContainer;
}

/**
 * Returns a u2d_DisplaySet useful for drawing to OpenGL the given mesh.
 * This function will prepare an object to show the boundary part of the
 * given mesh.
 * To avoid having to build the connectivity of boundary nodes with boundary
 * nodes numbering (we have them numbered along with domain elements) we pass
 * in all the mesh nodes and then we pass just the connectivity of boundary
 * elements in domain numbering. In this case the code is very similar to the
 * domain code, but with a different (reduced) connectivity passed in.
 * @param mesh the mesh object you want to draw.
 * @return A display set for drawing the given mesh.
 */
u2d_DisplaySet *u2d_DisplaySet::FromMeshBoundary(u2d_Mesh* mesh) {
     u2d_DisplaySet *displaySet = new u2d_DisplaySet();
    
    /* Vertexes */
    displaySet->setCoordPerVertex(mesh->getSpatialDimension());
    
    for(size_t i = 0; i < mesh->getDomainNodeNumber(); i++) {
        displaySet->pushVertexCoord(mesh->rr(0, i), mesh->rr(1, i));
    }
    
    
    /* Connectivity */
    displaySet->setVertexPerElement(2)
            .setVertexType(GL_LINES);
    
    for(size_t i = 0; i < mesh->getBoundaryElementNumber(); i++) {
        displaySet->pushTopologyElement(mesh->jjb(0, i) - 1,
                mesh->jjb(1, i) - 1);
    }
    
    return displaySet;     
}

u2d_DisplayContainer *u2d_DisplaySet::CreateKey(u2d_Mesh* mesh, u2d_Problem *problem) {
    u2d_DisplaySet *displaySet = new u2d_DisplaySet("KeyTicks");
    
    GLfloat keyOriginX;
    GLfloat keyOriginY;
    GLfloat width;
    GLfloat height;
    short int nTicks;
    GLfloat tickSpacing;
    float minX;
    float maxX;
    float minY;
    float maxY;
    
    /* Guess a correct origin */
    minX = mtl::min(mesh->rr[0][mtl::irange()]);
    maxX = mtl::max(mesh->rr[0][mtl::irange()]);
    minY = mtl::min(mesh->rr[1][mtl::irange()]);
    maxY = mtl::max(mesh->rr[1][mtl::irange()]);
    keyOriginX = maxX * 1.15;
    keyOriginY = minY;
    /* Guess width and height */
    width = (maxX - minX + maxY - minY) / 2 * 0.05; // 10% of average X and Y delta
    height = maxY - minY;
    /* Spacing and scale */
    nTicks = 20;
    tickSpacing = height/nTicks;

    displaySet->setCoordPerVertex(2)
            .setColorsPerVertex(3)
            .setVertexPerElement(4)
            .setVertexType(GL_LINES)
            .setFillMode(GL_LINE)
            .setLineColor(1.0, 1.0, 1.0);

    
    /* Draw the contour */
    /* Bottom */
    displaySet->pushVertexCoord(keyOriginX, keyOriginY);
    displaySet->pushVertexCoord(keyOriginX + width, keyOriginY);
    /* Left */
    displaySet->pushVertexCoord(keyOriginX, keyOriginY);
    displaySet->pushVertexCoord(keyOriginX, keyOriginY + tickSpacing * nTicks);
    /* Right */
    displaySet->pushVertexCoord(keyOriginX + width, keyOriginY);
    displaySet->pushVertexCoord(keyOriginX + width, keyOriginY + tickSpacing * nTicks);
    /* Top */
    displaySet->pushVertexCoord(keyOriginX, keyOriginY + tickSpacing * nTicks);
    displaySet->pushVertexCoord(keyOriginX + width, keyOriginY + tickSpacing * nTicks);
        
    /* Draw the ticks */
    for(short int i = 0; i <= nTicks; i++) {
        /* Lower vertexes */
        displaySet->pushVertexCoord(keyOriginX + 1.0 * width , keyOriginY + tickSpacing * i);
        displaySet->pushVertexCoord(keyOriginX + 1.2 * width, keyOriginY + tickSpacing * i);
    }
    
    /* Create the container and add the children */
    u2d_DisplayContainer *displayContainer = new u2d_DisplayContainer("Key");
    /* Now generate the gradient and add it as a child */
    displayContainer->addChild(CreateKeyGradient(mesh, problem));
    /* Add the key */
    displayContainer->addChild(displaySet);
    /* Create the tick labels and add them */
    float solutionMin = mtl::min(problem->getSolution());
    float solutionMax = mtl::max(problem->getSolution());
    float solutionDelta = solutionMax - solutionMin;
    
    for (short int i = 0; i <= nTicks; i++) {
        /* Lower vertexes */
        std::string label("KeyTickLabel");
        u2d_DisplayText *displayText = new u2d_DisplayText(label);
        displayText->setPosition(keyOriginX + 1.2 * width, keyOriginY + tickSpacing * i)
                .setSize(tickSpacing/1.5)
                .setOffset(30, -20);
        
        float tickValue = solutionMin + solutionDelta * i / nTicks;
        (*displayText) << tickValue;
        
        displayContainer->addChild(displayText);
        
    }
    
    return displayContainer;
    
}

u2d_DisplaySet *u2d_DisplaySet::CreateKeyGradient(u2d_Mesh* mesh, u2d_Problem *problem) {
    u2d_DisplaySet *displaySet = new u2d_DisplaySet("KeyGradient");
    
    GLfloat keyOriginX;
    GLfloat keyOriginY;
    GLfloat width;
    GLfloat height;
    short int nTicks;
    GLfloat tickSpacing;
    GLfloat color[3];
    float minX;
    float maxX;
    float minY;
    float maxY;
    
    /* Guess a correct origin */
    minX = mtl::min(mesh->rr[0][mtl::irange()]);
    maxX = mtl::max(mesh->rr[0][mtl::irange()]);
    minY = mtl::min(mesh->rr[1][mtl::irange()]);
    maxY = mtl::max(mesh->rr[1][mtl::irange()]);
    keyOriginX = maxX * 1.15;
    keyOriginY = minY;
    /* Guess width and height */
    width = (maxX - minX + maxY - minY) / 2 * 0.05; // 10% of average X and Y delta
    height = maxY - minY;
    /* Spacing and scale */
    nTicks = 20;
    tickSpacing = height/nTicks;

    displaySet->setCoordPerVertex(2)
            .setColorsPerVertex(3)
            .setVertexPerElement(4)
            .setVertexType(GL_QUADS)
            .setFillMode(GL_FILL);

    
    /* Draw the gradient */
    for(short int i = 0; i < nTicks; i++) {
        /* Lower vertexes */
        displaySet->pushVertexCoord(keyOriginX, keyOriginY + tickSpacing * i);
        displaySet->pushVertexCoord(keyOriginX + width, keyOriginY + tickSpacing * i);
        /* Lower vertexes colors */
        u2d_Math::LinearRGB(color, i, 0, 20);
        u2d_Math::LinearRGB(color, i, 0, 20);
        displaySet->pushVertexColor(color[0], color[1], color[2]);
        displaySet->pushVertexColor(color[0], color[1], color[2]);
        
        /* Upper vertexes */
        displaySet->pushVertexCoord(keyOriginX + width, keyOriginY + tickSpacing * (i + 1));
        displaySet->pushVertexCoord(keyOriginX, keyOriginY + tickSpacing * (i + 1));
        /* Upper vertexes colors */
        u2d_Math::LinearRGB(color, i + 1, 0, 20);
        u2d_Math::LinearRGB(color, i + 1, 0, 20);
        displaySet->pushVertexColor(color[0], color[1], color[2]);
        displaySet->pushVertexColor(color[0], color[1], color[2]);
        //std::cout << i + 1 << " R " << color[0] << "\tG " << color[1] << "\tB " << color[2] << std::endl;
    }
    
    return displaySet;
    
}

void u2d_DisplaySet::dump() const {
    std::cout << "Vertex array:" << std::endl;
    for(int i = 0; i < getVertexCoordCount(); i++) {
        std::cout << vertexArray[i] << std::endl;
    }
    
    std::cout << "Topology array:" << std::endl;
    for(int i = 0; i < getTopologyElementCount(); i++) {
        std::cout << topologyArray[i] << std::endl;
    }
    
    std::cout << "Color array:" << std::endl;
    for(int i = 0; i < getColorValuesCount(); i++) {
        std::cout << colorArray[i] << std::endl;
    }
    
    
}