/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2011 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on June 9, 2011, 11:05 PM
 */

#include <FTGL/ftgl.h>
#include <GL/gl.h>
#include <sstream>
#include <boost/shared_ptr.hpp>

#include "u2d_DisplayText.h"
#include "exceptions.h"

u2d_DisplayText::u2d_DisplayText(std::string _name /* = "" */)  :
fontSize(100),
name(_name),
enabled(true),
scaleFactor({1.0, 1.0, 1.0}),
position({0.0, 0.0, 0.0}),/* Unscaled */ // Initialization list for array is C++0x
color({1.0, 1.0, 1.0, 1.0}) /* White color */
{
    
    std::string fontName = "/usr/share/fonts/truetype/freefont/FreeSans.ttf";
    //std::string fontName = "/usr/share/fonts/truetype/freefont/FreeMono.ttf";
    //std::string fontName = "/usr/share/fonts/truetype/freefont/FreeSerif.ttf";
            

    /*
     * This font class is very fast when rendering a lot of text, since it
     * reuses already created textures when the same glyph use used multiple
     * times. Sadly it allocates ALL the possible glyphs, even those not used
     * and so takes up about 250 MiB of memory for each instance. So in our
     * case for each instance of u2d_DisplayText, since we are not reusing
     * the font class.
     */
    //this->font = boost::shared_ptr<FTFont>(new FTTextureFont(fontName.c_str()));
    
    /*
     * This font class is slower than FTTextureFont if the glyphs have to be
     * repeated more than once. On the other end, since weare creating a new
     * font instance for EACH u2d_DisplayText class we are forced to use
     * FTBufferFont. Indeed while FTBufferFont allocates memory just for the
     * text rendered, FTTextureFont allocates memory for ALL the possible glyphs
     * (i.e. about 250 MiB) and then places the glyphs.
     */
    this->font = boost::shared_ptr<FTFont>(new FTBufferFont(fontName.c_str()));
    
    // Those fonts are in window coordinates, they do not move, they are printed
    // where you give the position.
    //this->font = boost::shared_ptr<FTFont>(new FTBitmapFont(fontName.c_str()));
    //this->font = boost::shared_ptr<FTFont>(new FTPixmapFont(fontName.c_str()));

    //this->font = boost::shared_ptr<FTFont>(new FTPolygonFont(fontName.c_str()));
    //this->font = boost::shared_ptr<FTFont>(new FTOutlineFont(fontName.c_str()));
    
    if(this->font->Error()) {
        throw u2d_IO_Exception("Failed to load a suitable font");
    }
    
    /* Set font to a default size, size 100 and 100 dpi */
    this->font->FaceSize(fontSize, 100);
    
    /* Starting rendering position is (0.0, 0.0, 0.0) */
    this->offset = boost::shared_ptr<FTPoint>(new FTPoint(0.0, 0.0, 0.0));
    
}


u2d_DisplayText::~u2d_DisplayText() {

}

void u2d_DisplayText::draw() {
    if (isEnabled()) {
        glPushMatrix();
        glTranslatef(position[0], position[1], position[2]);
        glScalef(scaleFactor[0], scaleFactor[1], scaleFactor[2]);
        //glColor3f(1.0, 1.0, 1.0);
        glColor4f(color[0], color[1], color[2], color[3]);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        this->font->Render(this->text.str().c_str(), -1, *(this->offset));
        //std::cout << "Draw: " << this->getName() << std::endl;
        glPopMatrix();
    }
}

u2d_DisplayText &u2d_DisplayText::setText(std::string _text) {
    this->text.str(_text);
    return *this;
}

u2d_DisplayText &u2d_DisplayText::setFontSize(GLushort size, GLushort dpi /* = 100 */) {
    this->fontSize = size;
    this->font->FaceSize(size, dpi);
    return *this;
}

/**
 * Scale the font to the given OpenGL size.
 * This will scale the font to the given size. Notice that this will also
 * affect the font offset, since the scaling will affect the coordinate
 * position that are given with setOffset. To keep a consistent behavior
 * always setSize _before_ setOffset.
 * @param _size The size of the font in OpenGL units.
 * @return 
 */
u2d_DisplayText &u2d_DisplayText::setSize(GLfloat _size) {
    GLfloat _scaleFactor = _size/fontSize;
    this->scaleFactor[0] = _scaleFactor;
    this->scaleFactor[1] = _scaleFactor;
    this->scaleFactor[2] = 1.0;
    return *this;
}

/**
 * Sets the position where to render the font in OpenGL coordinates.
 * @param x
 * @param y
 * @param z
 * @return 
 */
u2d_DisplayText &u2d_DisplayText::setPosition(GLfloat x, GLfloat y, GLfloat z /* = 0.0 */) {
    position[0] = x;
    position[1] = y;
    position[2] = z;
    return *this;
}

/**
 * Sets an offset where to render the font. The offset is given in percentage
 * of the font size. e.g.
 * setOffset(0, -100) will move the font downwards of one full font size.
 * This function is very useful for fine font positioning, like if you want
 * to have a font that is in the middle of a given  point. To obtain this you
 * may specify the position with setPosition and offset the font then by
 * the wanted amount.
 * @rematk This is influenced by setSize and must be called always after
 * setSize has been called. This will result in a consistent behavior.
 * @param x
 * @param y
 * @param z
 * @return 
 */
u2d_DisplayText &u2d_DisplayText::setOffset(GLfloat x, GLfloat y, GLfloat z /* = 0.0 */) {
    this->offset.reset();
    this->offset = boost::shared_ptr<FTPoint>(new FTPoint(x, y, z));
    return *this;
}

u2d_DisplayText &u2d_DisplayText::setColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a /* = 1.0 */) {
    this->color[0] = r;
    this->color[1] = g;
    this->color[2] = b;
    this->color[3] = a;
    return *this;
}
    
std::string u2d_DisplayText::getName() const {
    return this->name;
}

u2d_DisplayText& u2d_DisplayText::enable() {
    this->enabled = true;
    return *this;
}

u2d_DisplayText& u2d_DisplayText::disable() {
    this->enabled = false;
    return *this;
}

bool u2d_DisplayText::isEnabled() const {
    return enabled;
}