/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on April 16, 2010, 4:31 PM
 */

#include "boost/numeric/mtl/mtl.hpp"
#include "u2d_load_scalarfield.h"
#include "exceptions.h"

u2d_LoadScalarField::u2d_LoadScalarField(string _filename, bool _lineonesize, size_t _suggested_size) :
u2d_LoadValues(_filename, _lineonesize, _suggested_size),
read_vector(0, 0.0),
vector_size(0) {
    parseValuesFile();
}

u2d_LoadScalarField::~u2d_LoadScalarField() {
}

mtl::dense_vector<double> u2d_LoadScalarField::getVector() {
    return read_vector;
}

void u2d_LoadScalarField::parseValuesFile()
throw (u2d_Parse_Error_Exception &) {
    if (lineonesize) {
        parseSize();
    }
    parseElements();
}
/**
 * Parse the first line of the file counting the size.
 * The size should be in the form of an integer number on the first valid
 * line (not starting with an #) of the file. i.e.:
 * # A scalar filed
 * # size
 * 345
 * # The data evaluated on nodes
 * 455.76
 * 45.78
 * ....
 * ....
 * # the above was last element, that is 345th
 * # any line of comment can be here or empty line
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_LoadScalarField::parseSize()
throw (u2d_Parse_Error_Exception &) {
    string line;
    bool found = false;
    std::cout << "--== Reading vector size  ==--" << std::endl;
    while (!found && getline(filestream, line).good()) {
        std::string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if (idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            sscanf(line.c_str(), "%zu", &vector_size);
            found = true;
        }
    }
    if (found == false) {
        throw u2d_Parse_Error_Exception("Cannot find vector size");
    }
    else if (suggested_size > 0 && vector_size != suggested_size) {
        // TODO fix missing sizes
        std::string errormsg = "Parsed vector size is: "; //+ vector_size;
        std::string errormsg2 = " but suggested size was: "; //+ suggested_size;
        throw u2d_Parse_Error_Exception(errormsg + errormsg2);
    }
    std::cout << "Vector elements: " << vector_size << std::endl;
    if(suggested_size != 0) {
        std::cout << "Vector elements (suggested): " << suggested_size << std::endl;
    }
    std::cout << "--== Done ==--" << std::endl;
}


void u2d_LoadScalarField::parseElements()
throw (u2d_Parse_Error_Exception &) {
    std::string line;
    std::vector<double> values;
    if(vector_size > 0) {
        values.reserve(vector_size);
    }
    std::cout << "--== Reading vector elements  ==--" << std::endl;
    while (getline(filestream, line).good()) {
        double a = 0.0;
        std::string::size_type idx = line.find_first_not_of(" \t\n\v");
        if (idx != line.npos && line[idx] == '#') {
            /* Line was starting with a # */
        } else if(idx == line.npos) {
            /* Line was empty */
        } else {
            /* Now we found the first good line, so we scan it and exit */
            sscanf(line.c_str(), "%lE", &a);
            values.push_back(a);
        }
    }
    // TODO Checking if the exit was for error or eof and throw if needed
    if (suggested_size > 0 && values.size() != suggested_size) {
        // TODO fix missing sizes
        std::string errormsg = "Parsed element number is: "; //+ values.size();
        std::string errormsg2 = " but suggested size was: "; //+ suggested_size;
        throw u2d_Parse_Error_Exception(errormsg + errormsg2);
    } else if(vector_size > 0 && vector_size != values.size()) {
        // TODO fix missing sizes
        std::string errormsg = "Parsed element number is: "; //+ values.size();
        std::string errormsg2 = " but size specified in file was: "; //+ vector_size;
        throw u2d_Parse_Error_Exception(errormsg + errormsg2);
    }
    /* Convert the vector to mtl vector */
    read_vector.change_dim(values.size());
    size_t index = 0;
    for (std::vector<double>::iterator it = values.begin(); it != values.end(); ++it) {
        read_vector[index] = *it;
        index++;
    }
    std::cout << "Read a vector of size: " << mtl::size(read_vector) << std::endl;
    std::cout << "--== Done ==--" << std::endl;
}
