/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on January 27, 2010, 4:55 PM
 */

#ifndef _U2D_PROBLEM_H
#define	_U2D_PROBLEM_H

#include "boost/numeric/mtl/mtl.hpp"


/**
 * This is a Abstract Base Class that holds a generic problem. Basically
 * a problem is the set of information needed as input to a solver. For
 * example for a linear system it consists of the matrix A and the right-hand
 * side vector b, in the expression \f[ Ax = b \f].
 */
class u2d_Problem {
public:
    //u2d_Problem(int _size);
    //u2d_Problem(const u2d_Problem& orig);
    //virtual ~u2d_Problem();
    virtual void addTerm(int i, int j, double val) = 0;
    virtual void setTerm(int i, int j, double val) = 0;
    virtual int getSize() = 0;
    virtual void addRHSTerm(int i, double val) = 0;
    virtual void setRHSTerm(int i, double val) = 0;
    virtual mtl::dense_vector<double> getSolution() = 0;
};

/**
 * This is a template for linear problems that should work with all the matrix
 * formats supported by MTL4. It is just a easy way of generating a subclass
 * of u2d_Problem with the right matrix type in it. The most basic type is
 * naturally one with a dense2D matrix type like:
 *  u2d_LinearProblem<dense2D<double> >.
 */
template <typename Matrix>
class u2d_LinearProblem : public u2d_Problem {
public:
    u2d_LinearProblem(int _size);
    u2d_LinearProblem(const u2d_LinearProblem& orig);
    virtual ~u2d_LinearProblem();
    void addTerm(int i, int j, double val);
    void setTerm(int i, int j, double val);
    void addRHSTerm(int i, double val);
    void setRHSTerm(int i, double val);
    mtl::dense_vector<double> getSolution();
    int getSize();
    bool getReady();
    bool unReady();

    /* Matrix, rhs and solution */
    Matrix A;
    mtl::dense_vector<double> b;
    mtl::dense_vector<double> x;

    /* PUBLIC FOR DEBUGGING */
    mtl::matrix::inserter<Matrix> *setter;
    mtl::matrix::inserter<Matrix, mtl::update_plus<double> > *adder;
private:
    int size;
};

template <typename Matrix>
void u2d_LinearProblem<Matrix>::setTerm(int i, int j, double val) {

    // Add value in m[i][j]
    (*setter)[i][j] << val;
    // Destructor of ins sets final state of m
}

template <typename Matrix>
void u2d_LinearProblem<Matrix>::addTerm(int i, int j, double val) {

    // Insert value in m[i][j]
    (*adder)[i][j] << val;
    // Destructor of ins sets final state of m
}

/**
 * Set the position <i> of the right hand side term to the passed value.
 * Set as in it's resetting it, that is:
 * b(534) -> 55
 * addRHSTerm(534, 10)
 * b(534) -> 10
 * @note @param <i> is _not_ the ID of the element, but it's the index.
 * notice that those can (and often do) differ for example of 1 if
 * vectors are zero-indicized and ID are one-indicized.
 */
template <typename Matrix>
void u2d_LinearProblem<Matrix>::setRHSTerm(int i, double val) {

    // Set value in b[i]
    b[i] = val;
}

/**
 * Adds the passed value to position <i> of the right hand side term
 * (the known term). Add as in it's incremental, that is:
 * b(534) -> 55
 * addRHSTerm(534, 10)
 * b(534) -> 65
 * @note @param <i> is _not_ the ID of the element, but it's the index.
 * notice that those can (and often do) differ for example of 1 if
 * vectors are zero-indicized and ID are one-indicized.
 */
template <typename Matrix>
void u2d_LinearProblem<Matrix>::addRHSTerm(int i, double val) {

    // Add value to b[i]
    b[i] += val;
}

template <typename Matrix>
u2d_LinearProblem<Matrix>::u2d_LinearProblem(int _size) :
A(_size, _size),
b(_size, 0.0),
x(_size, 0.0),
setter(NULL),
adder(NULL),
size(_size) {
    this->A = 0.0;
    this->setter = new mtl::matrix::inserter<Matrix>(A) ;
    this->adder = new mtl::matrix::inserter<Matrix, mtl::update_plus<double> >(A);
}

template <typename Matrix>
u2d_LinearProblem<Matrix>::u2d_LinearProblem(const u2d_LinearProblem& orig) {
}

template <typename Matrix>
u2d_LinearProblem<Matrix>::~u2d_LinearProblem() {
}

/**
 * Get the solution vector x. This is empty before solve() has been called
 * @return The solution vector x
 */
template <typename Matrix>
mtl::dense_vector<double> u2d_LinearProblem<Matrix>::getSolution() {
    return x;
}

/**
 * Gets the problem size as set with the constructor. The problem size is the
 * matrix (NXN) size N and the vector (NX1) length.
 * TODO: Change return type to std::size_t.
 * @return The problem size.
 */
template <typename Matrix>
int u2d_LinearProblem<Matrix>::getSize() {
    return size;
}

/**
 * This function prepares the problem to access terms of the A matrix. This
 * does not simply get the ready state of the problem, but also prepares it.
 * It should always be called before operating with the problem. After this has
 * been called you cannot anymore add or set the terms of the matric via
 * set*Term() or add*Term() functions. If after calling this you want to
 * start again to operate on the problem elements you must call unReady().
 * Notice that the problem starts unreadied and must be readied before being
 * resolved.
 * @remarks Currently this returns always true since there is no condition
 * in which the problem can be not readied, but this may change in the future.
 * @return true if the problem has become ready. False otherwise.
 */
template <typename Matrix>
bool u2d_LinearProblem<Matrix>::getReady() {
    if(this->setter != NULL) {
        delete this->setter;
        this->setter = NULL;
    }
    if(this->adder != NULL) {
        delete this->adder;
        this->adder = NULL;
    }
    return true;
}

/**
 * This functions unready the problem and allows set*Term() and add*Term()
 * functions to be used again. Calling getReady() normally disable the use of
 * set and add functions, calling this reneable it. Notice that the problem
 * starts unreadied.
 * @remarks Currently this returns always true since there is no condition
 * in which the problem can be not unreadied, but this may change in the future.
 * @return True if the problem has been unreadied again, false otherwise.
 */
template <typename Matrix>
bool u2d_LinearProblem<Matrix>::unReady() {
    if (this->setter == NULL) {
        this->setter = new mtl::matrix::inserter<Matrix > (A);
    }
    if (this->adder == NULL) {
        this->adder = new mtl::matrix::inserter<Matrix, mtl::update_plus<double> >(A);

    }
    return true;
}

#endif	/* _U2D_PROBLEM_H */

